/*
 * Author: Samy Al Bahra  <samy@kerneled.org>
 * Date  : Wed Feb 12 19:06:14 2003
 * Copyright(C) 2002-2003 Samy Al Bahra
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: config.c,v 1.5 2003/12/27 21:53:52 samy Exp $
 */

#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <config.h>
#include <headline.h>
#include <wstring.h>

/* It would be terrible to retype this every time... */
#define KERROR(x, y) \
	if (!isdigit(x) && !isalpha(x)) { \
		fprintf(stderr, "Parse error at line %d in configuration file, %s option requires" \
				"alphanumeric character argument\n", i, y); \
		return -1; \
	}

/* This function parses the wnews configuration file */
int
cconfig(void)
{
	char *line, *fname, *parse, *option, *set, *xset;
	unsigned int i = 0, kc;
	FILE *cp;

	if ((line = malloc(LMAX + 1)) == NULL) {
		fprintf(stderr, "Could not allocate memory for configuration file parsing\n");
		return -1;
	}
	if ((fname = malloc(LMAX + 1)) == NULL) {
		fprintf(stderr, "Could not allocate memory for file name buffer\n");
		return -1;
	}
	sprintf(fname, "%s%s", me->pw_dir, CONFIG);
	if ((cp = fopen(fname, "r")) == NULL) {
		fprintf(stderr, "Could not open configuration file for parsing (%s)\n", strerror(errno));
		return -1;
	}
	while (fgets(line, LMAX, cp)) {
		i++;
		parse = ltrim(line);
		if (parse[0] == '#' || parse[0] == 0)
			continue;
		option = strsep(&parse, " ");
		if (!option) {
			fprintf(stderr, "Parse error at line %d in configuration file, line must be in form <option> <value>\n", i);
			return -1;
		}
		xset = strsep(&parse, "\n");
		if (!xset) {
			fprintf(stderr, "Parse error at line %d in configuration file, line must be in form <option> <value>\n", i);
			return -1;
		}
		set = ltrim(xset);
		if (!set) {
			fprintf(stderr, "Parse error at line %d in configuration file, line must be in form <option> <value>\n", i);
			return -1;
		}
		if (!strcasecmp(option, PROXY)) {
			if (proxy) {
				fprintf(stderr, "Parse error at line %d in configuration file, proxy redefined\n", i);
				return -1;
			}
			proxy = strdup(rtrim(set));
		} else if (!strcasecmp(option, BROWSER)) {
			if (browser) {
				fprintf(stderr, "Parse error at line %d in configuration file, browser redefined\n", i);
				return -1;
			}
			if (!strstr(set, "%s")) {
				fprintf(stderr, "Parse error at line %d in configuration file, please use %%s in browser command\n", i);
				return -1;
			}
			if (strstr(set, "%n")) {
				fprintf(stderr, "Possible security risk at line %d in configuration file, please do not use %%n in browser command\n", i);
				return -1;
			}
			browser = strdup(set);
		} else if (!strcasecmp(option, GZIPO)) {
			if (gzip) {
				fprintf(stderr, "Parse error at line %d in configuration file, gzip redefined\n", i);
				return -1;
			}
			if (!strstr(set, "%s")) {
				fprintf(stderr, "Parse error at line %d in configuration file, please use %%s in gzip command\n", i);
				return -1;
			}
			if (strstr(set, "%n")) {
				fprintf(stderr, "Possible security risk at line %d in configuration file, please do not use %%n in gzip command", i);
				return -1;
			}
			gzip = strdup(set);
		} else if (!strcasecmp(option, TMPO)) {
			if (tmpdir) {
				fprintf(stderr, "Parser error at line %d in configuration file, temporary directory redefined\n", i);
				return -1;
			}
			if (strstr(set, "%n")) {
				fprintf(stderr, "Possible security risk at line %d in configuration file, please do not use %%n in browser command\n", i);
				return -1;
			}
			tmpdir = strdup(set);
		} else if (!strcasecmp(option, KUP)) {
			AUP = set[0];
			KERROR(AUP, KUP);
		} else if (!strcasecmp(option, KUD)) {
			AUD = set[0];
			KERROR(AUD, KUD);
		} else if (!strcasecmp(option, KURL)) {
			AURL = set[0];
			KERROR(AURL, KURL);
		} else if (!strcasecmp(option, KLEFT)) {
			ALEFT = set[0];
			KERROR(ALEFT, KLEFT);
		} else if (!strcasecmp(option, KRIGHT)) {
			ARIGHT = set[0];
			KERROR(ARIGHT, KRIGHT);
		} else if (!strcasecmp(option, KQUIT)) {
			AQUIT = set[0];
			KERROR(AQUIT, KQUIT);
		} else if (!strcasecmp(option, KBROWSER)) {
			ABROWSER = set[0];
			KERROR(ABROWSER, KBROWSER);
		} else if (!strcasecmp(option, KPUP)) {
			APUP = set[0];
			KERROR(APUP, KPUP);
		} else if (!strcasecmp(option, KPDN)) {
			APDN = set[0];
			KERROR(APDN, KPDN);
		} else if (!strcasecmp(option, KENTER)) {
			AENTER = set[0];
			KERROR(AENTER, KENTER);
		} else if (!strcasecmp(option, INTERACTIVE)) {
			if (!strcasecmp(rtrim(set), "yes"))
				interactive = 1;
			else
				interactive = 0;
		} else {
			fprintf(stderr, "Unknown option at line %d of configuration file\n", i);
			return -1;
		}
	}

	if (!tmpdir || !gzip) {
		fprintf(stderr, "%s and %s MUST be used in the wnews configuration file\n", TMPO, GZIPO);
		return -1;
	}

	return 0;
}

/* This function parses the resources configuration file
 * and returns the global resource structure based on this
 */
struct resource *
rconfig(void)
{
	char *line, *fname, *parse;
	unsigned int i = 0, x = 0, y;
	FILE *rp;
	struct resource *next, *final = (struct resource *) NULL;

	if (!(line = malloc(LMAX + 1))) {
		fprintf(stderr, "Could not allocate memory for resources file parsing\n");
		return (struct resource *) NULL;
	}
	if (!(fname = malloc(LMAX + 1))) {
		fprintf(stderr, "Could not allocate memory for file name buffer\n");
		return (struct resource *) NULL;
	}
	sprintf(fname, "%s%s", me->pw_dir, RSOURC);
	if ((rp = fopen(fname, "r")) == NULL) {
		fprintf(stderr, "Could not open resources file for parsing (%s)\n", strerror(errno));
		return (struct resource *) NULL;
	}
	while (fgets(line, LMAX, rp)) {
		i++;
		parse = ltrim(line);
		if (parse[0] == '#' || parse[0] == 0)
			continue;

		for (x = 0, y = 0; x <= strlen(parse); x++) {
			if (parse[x] == '|')
				y++;
		}

		if (y != 2) {
			fprintf(stderr, "Parse error at line %d in resources file, must be in form <Title>|<Description>|<Address>\n", i);
			return (struct resource *) NULL;
		}
		if ((next = malloc(sizeof(struct resource))) == NULL) {
			fprintf(stderr, "Could not allocate memory for resources structure construction\n");
			return (struct resource *) NULL;
		}
		next->title = strdup(strsep(&parse, "|"));
		next->desc = strdup(strsep(&parse, "|"));
		next->address = strdup(strsep(&parse, "\n"));
		next->next = (struct resource *) NULL;

		if (final)
			next->next = final;

		final = next;
		numr++;
	}

	free(fname);
	free(line);
	fclose(rp);
	return final;
}

/* This file checks for the availability of various configuration
 * files and creates them if need be
 */
int
config(void)
{
	char *junk;
	FILE *cfg;
	struct stat file;

	if (!(me = getpwuid(getuid()))) {
		fprintf(stderr, "Could not get password file entry for user\n");
		return -1;
	}
	if (!(junk = malloc(LMAX + strlen(RSOURC + 1)))) {
		fprintf(stderr, "Could not allocate memory for junk buffer\n");
		return -1;
	}
	sprintf(junk, "%s%s", me->pw_dir, DIRDEF);
	if ((cfg = fopen(junk, "r")) == NULL) {
		fprintf(stderr, "Could not open wnews configuration directory (%s)\n", strerror(errno));
		if (errno != ENOENT)
			return -1;
		fprintf(stderr, "Creating directory...\n");
		mkdir(junk, S_IRUSR | S_IWUSR | S_IXUSR);
		if ((cfg = fopen(junk, "r")) == NULL)
			return -1;
	}
	fstat(fileno(cfg), &file);
	if (!(file.st_rdev | S_IFDIR)) {
		fprintf(stderr, "Make sure that %s is a directory\n", junk);
		return -1;
	}
	fclose(cfg);

	sprintf(junk, "%s%s", me->pw_dir, CONFIG);
	if (!(cfg = fopen(junk, "r"))) {
		fprintf(stderr, "Could not open wnews configuration file (%s)\n", strerror(errno));
		if (errno != ENOENT)
			return -1;
		fprintf(stderr, "Creating configuration file...\n");
		if (!(cfg = fopen(junk, "w"))) {
			fprintf(stderr, "Could not create wnews configuration file (%s)\n", strerror(errno));
			return -1;
		}
		fprintf(cfg, "%s", df_config);
		fclose(cfg);
	}
	sprintf(junk, "%s%s", me->pw_dir, RSOURC);
	if (!(cfg = fopen(junk, "r"))) {
		fprintf(stderr, "Could not open wnews resources file (%s)\n", strerror(errno));
		if (errno != ENOENT)
			return -1;
		fprintf(stderr, "Creating wnews resources file...\n");
		if (!(cfg = fopen(junk, "w"))) {
			fprintf(stderr, "Could not create wnews resources file (%s)\n", strerror(errno));
			return -1;
		}
		fprintf(cfg, "%s", df_resources);
		fclose(cfg);
	}
	free(junk);
	return 0;
}
