/*
 * Author: Mohammad Susli  <abionnnn@kerneled.com>
 * Date  : Tue Dec 31 02:23:40 2002
 * Copyright(C) 2002-2003 Mohamad Susli
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: rss.c,v 1.3 2003/12/27 21:54:15 samy Exp $
 */

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <config.h>
#include <error.h>
#include <headline.h>
#include <wstring.h>

/* The order of these match the order of the list. */
#define	HEADLINE	0x00
#define	LINK		0x01
#define	DESC		0x02
char *tag[3] = {"title", "link", "description"};
char *etag[3] = {"/title", "/link", "/description"};

#define ORIENT_END		0x00
#define ORIENT_BEGIN	0x01

struct content *
new_rss_node(char *h, char *d, char *l)
{
	struct content *nn;

	if (!(nn = malloc(sizeof(struct content)))) {
		werror("Could not allocate memory for new rss feed");
		return NULL;
	}
	if (!h || !d || !l)
		return NULL;

	nn->headline = h;
	nn->desc = d;
	nn->link = l;
	nn->next = NULL;

	return nn;
}

char *
find_tag(char *in, char *tag, int orient)
{
	char *cur;
	char *b, *e;
	unsigned char found = 0, q = 0, bfound = 0, efound = 0;
	char c;

	if (in && *in)
		cur = in;
	else
		return in;

	while (!found) {
		while (!found && *cur) {
			c = *cur;
//			if (c == '\"')
//				q = q ? 0 : 1;
//			if (!q) {
				switch (c) {
				case '<':
					bfound = 1;
					b = cur;
					break;
				case '>':
					efound = 1;
					e = cur;
					break;
				}
				if (efound && !bfound)
					efound = 0;	/* reject > outliers */
				if (efound && bfound)
					found = 1;
//			}
			++cur;
		}
		if (!found)
			return NULL;	/* end of string? */
		if (strncmp(tag, b + 1, strlen(tag)) != 0) {
			found = 0;
			efound = 0;
			bfound = 0;
		}
	}

	if (orient == ORIENT_END)
		return b - 1;
	else if (orient == ORIENT_BEGIN)
		return e + 1;

	return e + 1;
}


struct content *
getrss(const char *buffer)
{
	struct content *root = NULL, *cur_node = NULL, *nnode;
	char *cur_b = (char *)buffer;
	char *s, *end, *e;
	char *hay;
	char *tagi[3];
	int i;

	numc = 0;
	while (1) {
		s = find_tag(cur_b, "item", ORIENT_BEGIN);
		if (!s)
			return root;
		end = find_tag(s, "/item", ORIENT_END);
		if (!end)
			return root;
		hay = strgrab(s, end);
		
		if (!hay) {
			werror("RSS parser has failed");
			return NULL;
		}
			
		
		cur_b = end;

		/* Grab what we need quick 'n dirty. [add error checking.] */
		for (i = 0; i < 3; i++) {
			s = find_tag(hay, tag[i], ORIENT_BEGIN);
			e = find_tag(hay, etag[i], ORIENT_END);

			if (s) {
				tagi[i] = strgrab(s, e);
				if (!tagi[i]) {
					werror("Failed to parse RSS file");
					return NULL;
				}
			} else
				tagi[i] = strdup("");
		}

		if (*(tagi[0]) == '\0' && *(tagi[1]) == '\0' && *(tagi[2]) == '\0') {
			for (i=0;i<3;i++) free(tagi[i]);
			free(hay);
			continue;
		}

		if (cur_node != NULL) {
			nnode = new_rss_node(tagi[HEADLINE], tagi[DESC], tagi[LINK]);
			cur_node->next = nnode;
			cur_node = nnode;
		} else {
			root = new_rss_node(tagi[HEADLINE], tagi[DESC], tagi[LINK]);
			cur_node = root;
		}

		numc++;
		free(hay);
	}

	return root;
}
