/*
 * Author: Mohammad Susli <abionnnn@kerneled.com>
 * Date: Thur Jan 03:06:20 2003
 * Copyright(C) 2002-2003 Mohammad Susli
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: socket.c,v 1.5 2003/12/27 21:54:34 samy Exp $
 */

#include <curses.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <netdb.h>
#include <resolv.h>
#include <unistd.h>

#include <config.h>
#include <headline.h>
#include <interface.h>
#include <wstring.h>
#include <socket.h>

#define HTTP_PORT	80
#define PROXY_PORT	8080
#define	BUFINC		1460
#define	BUFTOL		100
#define	TYPE_CRLF	0
#define	TYPE_CR		1
#define	TYPE_LF		2

const char *cr_keys[3] = {"\r\n", "\n", "\r"};

#define HTTP_STRING			"GET %s HTTP/1.1\r\n"			 \
					"Host: %s:%s\nUser-Agent: wnews/0.1\r\n" \
					"Accept-Encoding: gzip\r\n"		 \
					"Connection: close\r\n"			 \
					"\r\n"

#define	TAG_CONTENT_ENCODING		"Content-Encoding"
#define	TAG_CONTENT_LENGTH		"Content-Length"

#define	SLEEP_OUT			10000000

inline int 
select_read(int sock)
{
	fd_set rfds;
	struct timeval tv;

	FD_ZERO(&rfds);
	FD_SET(sock, &rfds);

	tv.tv_sec = 0;
	tv.tv_usec = SLEEP_OUT;

	select(sock + 1, &rfds, NULL, NULL, &tv);

	return FD_ISSET(sock, &rfds);
}

inline int 
select_write(int sock)
{
	fd_set wfds;
	struct timeval tv;

	FD_ZERO(&wfds);
	FD_SET(sock, &wfds);

	tv.tv_sec = 0;
	tv.tv_usec = SLEEP_OUT;

	select(sock + 1, NULL, &wfds, NULL, &tv);

	return FD_ISSET(sock, &wfds);
}


/* Sends all of buf to fd */
int 
sendall(int fd, char *buf)
{
	char *s = buf;
	int len = strlen(buf);
	int r, t = 0;

	while (t != len && select_write(fd)) {
		if (!select_write(fd)) {
			werror("Network code timed out on write");
		}
		r = send(fd, s, len - t, 0);

		if (r == -1)
			return 0;

		t += r;
		s += r;
	}

	return -1;
}

/* Will only accept a list of (char *)'s */
int 
sendfc(int fd, char *fmt,...)
{
	va_list ap;
	char *buf, *s;
	int i, n = 0, len = 0;

	if (!fmt)
		return 0;

	s = strstr(fmt, "%s");

	while (s) {
		s += 2;
		++n;
		s = strstr(s, "%s");
	}

	va_start(ap, fmt);
	for (i = 0; i < n; i++) {
		s = va_arg(ap, char *);

		if (!s)
			return 0;

		len += strlen(s);
	}
	va_end(ap);
	len += strlen(fmt) - 2 * n + 1;

	va_start(ap, fmt);
	buf = malloc(len);
	vsprintf(buf, fmt, ap);

	if (!sendall(fd, buf)) {
		free(buf);
		return 0;
	}
	free(buf);

	return -1;
}

/* Last resort ... recieve everything until the server disconnects. */
char *
recieveall(int fd)
{
	char *buf;
	char *s;
	int r, size, left;

	size = BUFINC;
	left = size;

	buf = malloc(size);
	if (!buf) {
		werror("Could not allocate memory for network receive buffer");
		return NULL;
	}
	s = buf;
	r = -1;

	while (r) {
		if (!select_read(fd)) {
			werror("Network connection to HTTP server timed out");
			free(buf);
			return NULL;
		}
		r = recv(fd, s, left, 0);
		if (r == -1) {
			werror("Network error, could not recieve data from HTTP server");
			free(buf);
			return NULL;
		}
		left -= r;
		s += r;

		if (interactive)
			wstat("Downloading... %d bytes", (size - left));

		/* does our buffer need toping up? =) */
		if (left < BUFTOL) {
			left += BUFINC;
			size += BUFINC;
			if (!(buf = realloc(buf, size))) {
				werror("Could not make recv buffer grow");
				return NULL;
			}
			s = buf + (size - left);
		}
	}

	if (!(buf = realloc(buf, size - left + 1))) {
		werror("Could not resize the recv buffer");
		return NULL;
	}
	buf[size - left] = '\0';

	return buf;
}

/*
 * Recieves raw bytes and saves them into memory. caps the end with '\0'
 * (this shouldn't effect gzip since we just ignore that end byte leaving it
 * for the OS to free up the memory later anyway.
 */
char *
recvn(int fd, int num)
{
	char *buf, *s;
	int r, left;

	buf = malloc(num + 1);
	if (!buf) {
		werror("Could not allocate memory for network receive buffer");
		return NULL;
	}
	s = buf;
	r = -1;
	left = num;

	while (left) {
		if (!select_read(fd)) {
			werror("Network connection to HTTP server timed out");
			free(buf);
			return NULL;
		}
		r = recv(fd, s, left, 0);
		if (r == -1) {
			werror("Network error, could not recieve data from HTTP server");
			free(buf);
			return NULL;
		}
		left -= r;
		s += r;

		if (interactive)
			wstat("Downloading... (%dK/%dK)", (num - left) >> 8, num >> 8);

		if (!r) {
			werror("Server disconnected.");
			free(buf);
			return NULL;
		}
	}

	buf[num] = '\0';
	return buf;
}

/* use only on ANSI connection. */
char *
recvto(int fd, const char *t)
{
	char *buf;
	char *e;
	int size, nl, r;

	size = BUFINC;

	buf = malloc(size + 1);
	r = -1;

	while (r) {
		if (!select_read(fd)) {
			werror("Connection Timed out.\n");
			free(buf);
			return NULL;
		}
		r = recv(fd, buf, size, MSG_PEEK);
		if (r == -1)
			continue;

		buf[size] = '\0';	/* so string search returns correct
					 * result. */

		if ((e = strstr(buf, t)) != NULL) {
			free(buf);
			nl = (int)(e - buf) + strlen(t);
			buf = recvn(fd, nl);
			return buf;
		} else {
			size += BUFTOL;
			buf = (char *)realloc(buf, size + 1);
			r = -1;
			continue;
		}
	}

	/* return what we had so far */
	fputs("Did not recieve expected token.\n", stderr);
	free(buf);
	return NULL;
}

int 
connect_host(char *domain, int port)
{
	struct sockaddr_in serv_addr;
	struct hostent *host;
	int sock;

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock == -1) {
		werror("Could not obtain a socket");
		return 0;
	}
	memset(&serv_addr, 0, sizeof(struct sockaddr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(port);

	host = gethostbyname(domain);
	if (host == NULL) {
		werror("Host name look up failed");
		return 0;
	}
	serv_addr.sin_addr.s_addr = *(in_addr_t *)host->h_addr_list[0];

	if (connect(sock, (struct sockaddr *) & serv_addr, sizeof(serv_addr)) == -1) {
		werror("Could not connect to host");
		return 0;
	}
	return sock;
}

char *
gettag(char *buf, const char *tag, int cr_type)
{
	char *e, *s = buf;

	while ((s = strstr(s, cr_keys[cr_type])) != NULL) {
		s += strlen(cr_keys[cr_type]);
		s = ltrim(s);
		if (strncasecmp(s, tag, strlen(tag)) == 0) {
			s += strlen(tag);
			s = strchr(s, ':') + 1;
			s = ltrim(s);
			e = strstr(s, cr_keys[cr_type]);
			if (e)
				s = strgrab(s, e - 1);
			else
				s = strdup(s);
			return s;
		}
	}

	return NULL;
}

/* Analyze entire recieve buffer after send including HTTP header */
char *
analyzehttp(char *buf)
{
	char *s, *t, *h, *u;
	int i, l, cr_type;

	if (!buf) {
		werror("analyzehttp(): no buffer recieved");
		return NULL;
	}
	for (i = 0, s = NULL; i < 3 && !s; i++) {
		s = strstr(buf, cr_keys[i]);
		cr_type = i;
	}

	if (s) {
		t = malloc(2 * strlen(cr_keys[cr_type]) + 1);
		strcpy(t, cr_keys[cr_type]);
		strcat(t, cr_keys[cr_type]);
		s = strstr(buf, t);
		free(t);

		if (!s) {
			werror("Server does not adhere to RFC2616");
			return NULL;
		}
		h = strgrab(buf, s - 1);
		t = gettag(h, TAG_CONTENT_ENCODING, cr_type);
		if (t && (strstr(t, "gzip") || strstr(t, "deflate") || strstr(t, "compress"))) {
			free(t);
			t = gettag(h, TAG_CONTENT_LENGTH, cr_type);
			if (!t) {
				werror("Content-Length tag not found");
				return NULL;
			}
			free(h);
			l = atoi(t);
			free(t);
			u = malloc(l);
			memcpy(u, s + strlen(cr_keys[cr_type]) * 2, l);
			t = strgunzip(u, l);
			free(u);
			u = t;
			if (!u) {
				werror("Decompression failure");
				return NULL;
			}
			return u;
		} else {
			if (t)
				free(t);
			free(h);
			u = strdup(s + strlen(cr_keys[cr_type]));
			return u;
		}
	}
	werror("Server does not adhere to RFC2616");
	return NULL;
}

/*
 * Recieve HTTP assuming already requested. Saves as buffer '\0' terminated.
 * Also if gzip'ed decompresses.
 */
char *
recvhttp(int fd)
{
	char *buf, *t;
	int len, cr_type = 0;
	unsigned char f_cl, f_cz;

	t = malloc(2 * strlen(cr_keys[cr_type]) + 1);
	strcpy(t, cr_keys[cr_type]);
	strcat(t, cr_keys[cr_type]);

	buf = recvto(fd, t);
//TODO:how can we find out which kind of line terminator it is ?
		free(t);

	if (!buf) {
		werror("Server does not adhere to RFC2616");
		return NULL;
	}
	t = gettag(buf, TAG_CONTENT_LENGTH, cr_type);
	if (!t) {
		f_cl = 0;
	} else {
		len = atoi(t);
		f_cl = 1;
		free(t);
	}

	t = gettag(buf, TAG_CONTENT_ENCODING, cr_type);
	if (t && (strstr(t, "gzip") || strstr(t, "deflate") || strstr(t, "compress"))) {
		free(t);
		f_cz = 1;
	} else {
		if (t)
			free(t);
		f_cz = 0;
	}
	free(buf);

	if (f_cz) {
		if (!f_cl) {
			werror("Server transfered compressed data without Content-length. bailing out.");
			return NULL;
		} //
		else {
			buf = recvn(fd, len);
			t = strgunzip(buf, len);
			free(buf);
			buf = t;
			//
		}
	} else {
		if (f_cl)
			buf = recvn(fd, len);
		else
			buf = recieveall(fd);
	}

	return buf;
}

/*
 * This function is huge because it is flexible; I will break it up into
 * pieces once we get compression working.
 */
char *
getpages(char *url, char *proxy)
{
	int sock;
	char *s, *h, *u, *t;
	char *buf;
	unsigned short port, proxy_port;

	if (!url)
		return NULL;
	s = strchr(url, '/');
	if (s) {
		h = strgrab(url, s - 1);
		u = strdup(s);
	} else {
		h = strdup(url);
		u = strdup("/");
	}

	s = strchr(h, ':');
	if (s && *(s + 1))
		port = (unsigned short)atoi(s + 1);
	else
		port = HTTP_PORT;

	if (proxy) {
		s = strchr(proxy, ':');
		if (s && *(s + 1)) {
			t = strgrab(proxy, s - 1);
			proxy_port = (unsigned short)atoi(s + 1);
		} else {
			t = proxy;
			proxy_port = PROXY_PORT;
		}

		if (!(sock = connect_host(t, proxy_port))) {
			free(t);
			free(h);
			return NULL;
		}
		/*
		 * if there is a error, it is because the proxy server does
		 * not accept : in the GET statement, which it *should*
		 */
		free(u);
		u = malloc(strlen(url) + 8);
		strcpy(u, "http://");
		u += 7;
		strcpy(u, url);
		u -= 7;
	} else {
		if (!(sock = connect_host(h, port))) {
			free(u);
			free(h);
			return NULL;
		}
	}

	s = malloc(8 * sizeof(int) + 1);
	sprintf(s, "%u", port);

	if (!sendfc(sock, HTTP_STRING, u, h, s)) {
		werror("Failed to send HTTP/1.1 request");
		free(s);
		free(h);
		return NULL;
	}
	/*
	 * buf = recieveall(sock); if (!buf) { werror("Could not retrieve
	 * data"); return NULL; } free(s); free(h); close(sock); t =
	 * analyzehttp(buf); free(buf);
	 */
	buf = recvhttp(sock);

	return buf;
}
